/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla Communicator client code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *     Laurent Jouanneau <laurent.jouanneau@disruptive-innovations.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsCOMPtr.h"
#include "nsIDOMNode.h"
#include "nsIDOMElement.h"
#include "nsIDOMNodeList.h"
#include "nsIServiceManager.h"
#include "nsUnicharUtils.h"
#include "mozIStorageDataSet.h"

#include "nsIURI.h"
#include "nsIIOService.h"
#include "nsIChannel.h"
#include "nsIFileChannel.h"
#include "nsIFile.h"

#include "nsXULContentUtils.h"
#include "nsXULTemplateBuilder.h"
#include "nsXULTemplateQueryProcessorStorage.h"
#include "nsXULTemplateResultStorage.h"

//----------------------------------------------------------------------
//
// nsXULTemplateResultSetStorage
//

NS_IMPL_ISUPPORTS1(nsXULTemplateResultSetStorage, nsISimpleEnumerator)

NS_IMETHODIMP
nsXULTemplateResultSetStorage::HasMoreElements(PRBool *aResult)
{
    return mStatement->ExecuteStep(aResult);
}

NS_IMETHODIMP
nsXULTemplateResultSetStorage::GetNext(nsISupports **aResult)
{
    nsXULTemplateResultStorage* result =
        new nsXULTemplateResultStorage(mProcessor, mStatement);
    if (! result)
        return NS_ERROR_OUT_OF_MEMORY;

    *aResult = result;
    NS_ADDREF(result);
    return NS_OK;
}



//----------------------------------------------------------------------
//
// nsXULTemplateQueryProcessorStorage
//

NS_IMPL_ISUPPORTS1(nsXULTemplateQueryProcessorStorage, nsIXULTemplateQueryProcessor)


nsXULTemplateQueryProcessorStorage::nsXULTemplateQueryProcessorStorage() 
    : mGenerationStarted(PR_FALSE)
{

}

nsresult
nsXULTemplateQueryProcessorStorage::InitStorage(const nsAString & aDatasourceName)
{

printf("InitStorage %s\n", NS_ConvertUTF16toUTF8(aDatasourceName).get());
    mStorage = do_GetService("@mozilla.org/storage/service;1");
    if (! mStorage)
        return NS_ERROR_FAILURE;

    if(aDatasourceName.EqualsLiteral("profile") || aDatasourceName.EqualsLiteral("memory")){
printf("InitStorage profile or memory\n");
        return mStorage->OpenSpecialDatabase(NS_ConvertUTF16toUTF8(aDatasourceName).get(),
                                        getter_AddRefs(mStorageConnection));
    }else{
printf("InitStorage file url\n");
        nsresult rv;
        nsCOMPtr<nsIChannel> channel;
        nsCOMPtr<nsIIOService> ioservice = do_GetService("@mozilla.org/network/io-service;1",&rv);
        NS_ENSURE_SUCCESS(rv,rv);

        rv = ioservice->NewChannel(NS_ConvertUTF16toUTF8(aDatasourceName),
                                   nsnull, nsnull, getter_AddRefs(channel));

        NS_ENSURE_SUCCESS(rv,rv);
        if(!channel)
            return NS_ERROR_NULL_POINTER;

        nsCOMPtr<nsIFileChannel> fileChannel = do_QueryInterface(channel);
        if(!fileChannel)
            return NS_ERROR_NULL_POINTER; // not a file url

        nsCOMPtr<nsIFile> file;
        rv = fileChannel->GetFile(getter_AddRefs(file));
        NS_ENSURE_SUCCESS(rv,rv);

printf("InitStorage ok, have a file, open\n");

        return mStorage->OpenDatabase(file, getter_AddRefs(mStorageConnection));
    }
}

PRInt32
nsXULTemplateQueryProcessorStorage::GetColumnIndex(nsIAtom* aColumnName)
{
    PRInt32 count = mColumnNames.Count();
    for (PRInt32 c = 0; c < count; c++) {
        if (mColumnNames[c] == aColumnName)
            return c;
    }

    return -1;
}

void
nsXULTemplateQueryProcessorStorage::FillColumnValues(mozIStorageValueArray* aRecord,
                                                     nsStringArray& aArray)
{
    PRInt32 count = mColumnNames.Count();
    for (PRInt32 c = 0; c < count; c++) {
        nsAutoString val;
        aRecord->GetString(c, val);
        aArray.AppendString(val);
    }
}

NS_IMETHODIMP
nsXULTemplateQueryProcessorStorage::InitializeForBuilding(nsISupports* aDatasource,
                                                          nsIXULTemplateBuilder* aBuilder,
                                                          nsIDOMNode* aRootNode)
{
printf("InitializeForBuilding\n");

    if (mGenerationStarted)
        return NS_ERROR_UNEXPECTED;

    nsAutoString datasource;
    nsresult rv;

    nsCOMPtr<nsIDOMElement> root = do_QueryInterface(aRootNode,&rv);
    NS_ENSURE_SUCCESS(rv,rv);

    rv = root->GetAttribute(NS_LITERAL_STRING("storage"), datasource);
    NS_ENSURE_SUCCESS(rv,rv);

    if(datasource.IsEmpty()){
        datasource.AssignASCII("profile");
    }

    rv = InitStorage(datasource);
    if(NS_FAILED(rv)){
printf("InitStorage failed\n");

        return NS_ERROR_INVALID_ARG;
    }
printf("InitStorage ok\n");
    return NS_OK;
}

NS_IMETHODIMP
nsXULTemplateQueryProcessorStorage::Done()
{
    mGenerationStarted = PR_FALSE;
    return NS_OK;
}

NS_IMETHODIMP
nsXULTemplateQueryProcessorStorage::CompileQuery(nsIXULTemplateBuilder* aBuilder,
                                                 nsIDOMNode* aQueryNode,
                                                 nsIAtom* aRefVariable,
                                                 nsIAtom* aMemberVariable,
                                                 nsISupports** _retval)
{
printf("CompileQuery\n");

    nsCOMPtr<nsIDOMNodeList> childNodes;
    aQueryNode->GetChildNodes(getter_AddRefs(childNodes));

    PRUint32 length;
    childNodes->GetLength(&length);

    nsCOMPtr<mozIStorageStatement> statement;

    for (PRUint32 n = 0; n < length; n++)
    {
        nsCOMPtr<nsIDOMNode> child;
        childNodes->Item(n, getter_AddRefs(child));

        PRUint16 nodeType;
        child->GetNodeType(&nodeType);

        if (nodeType == nsIDOMNode::TEXT_NODE) {
            nsAutoString content;
            child->GetNodeValue(content);

            nsresult rv = mStorageConnection->CreateStatement(NS_ConvertUTF16toUTF8(content),
                                                              getter_AddRefs(statement));
printf("CompileQuery found query\n");
            if (NS_FAILED(rv)) 
                return rv;
printf("CompileQuery query ok\n");
            break;
        }
    }

    // XXX this needs to be query specific

    PRUint32 count;
    statement->GetColumnCount(&count);

    for (PRUint32 c = 0; c < count; c++) {
        nsCAutoString name;
        statement->GetColumnName(c, name);
        mColumnNames.AppendObject(NS_NewAtom(NS_LITERAL_CSTRING("?") + name));
    }

    *_retval = statement;
    NS_IF_ADDREF(*_retval);

    return NS_OK;
}

NS_IMETHODIMP
nsXULTemplateQueryProcessorStorage::GenerateResults(nsISupports* aDatasource,
                                         nsIXULTemplateResult* aRef,
                                         nsISupports* aQuery,
                                         nsISimpleEnumerator** aResults)
{
printf("GenerateResults\n");
    mGenerationStarted = PR_TRUE;

    nsCOMPtr<mozIStorageStatement> statement = do_QueryInterface(aQuery);
    if (! statement)
        return NS_ERROR_FAILURE;

    nsXULTemplateResultSetStorage* results =
        new nsXULTemplateResultSetStorage(this, statement);
    if (! results)
        return NS_ERROR_OUT_OF_MEMORY;
printf("GenerateResults ok\n");
    *aResults = results;
    NS_ADDREF(*aResults);

    return NS_OK;
}

NS_IMETHODIMP
nsXULTemplateQueryProcessorStorage::AddBinding(nsIDOMNode* aRuleNode,
                                    nsIAtom* aVar,
                                    nsIAtom* aRef,
                                    const nsAString& aExpr)
{
    return NS_OK;
}

NS_IMETHODIMP nsXULTemplateQueryProcessorStorage::TranslateRef(nsISupports* aDatasource,
                                                    const nsAString& aRefString,
                                                    nsIXULTemplateResult** aRef)
{
    nsXULTemplateResultStorage* result =
        new nsXULTemplateResultStorage(this, nsnull);
    if (! result)
        return NS_ERROR_OUT_OF_MEMORY;

    *aRef = result;
    NS_ADDREF(*aRef);

    return NS_OK;
}


NS_IMETHODIMP nsXULTemplateQueryProcessorStorage::CompareResults(nsIXULTemplateResult* aLeft,
                                                      nsIXULTemplateResult* aRight,
                                                      nsIAtom* aVar,
                                                      PRInt32* aResult)
{
    *aResult = 0;

    // XXXndeakin handle other types

    nsAutoString leftVal;
    aLeft->GetBindingFor(aVar, leftVal);

    nsAutoString rightVal;
    aRight->GetBindingFor(aVar, rightVal);

    *aResult = ::Compare(nsDependentString(leftVal),
                         nsDependentString(rightVal),
                         nsCaseInsensitiveStringComparator());
    return NS_OK;
}
